# Written by retoor@molodetz.nl

# This code provides decorators for caching function results with expiration times, supporting both synchronous and asynchronous functions.

# Imports: `time`, `asyncio`, and `wraps` from `functools` are used in this script.

# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json
import time
from functools import wraps


def time_cache(timeout: int = 600):
    def decorator(func):
        cache = {}

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (
                json.dumps(args, default=str),
                json.dumps(frozenset(kwargs.items()), default=str),
            )
            current_time = time.time()

            if key in cache:
                result, timestamp = cache[key]
                if current_time - timestamp < timeout:
                    return result

            result = func(*args, **kwargs)
            cache[key] = (result, current_time)
            return result

        return wrapper

    return decorator


def time_cache_async(timeout: int = 600):
    def decorator(func):
        cache = {}

        @wraps(func)
        async def wrapper(*args, **kwargs):
            key = (
                json.dumps(args, default=str),
                json.dumps(frozenset(kwargs.items()), default=str),
            )
            current_time = time.time()

            if key in cache:
                result, timestamp = cache[key]
                if current_time - timestamp < timeout:
                    return result

            result = await func(*args, **kwargs)
            cache[key] = (result, current_time)
            return result

        return wrapper

    return decorator
