from datetime import datetime
from functools import partial
from xmlrpc.client import (
    Fault,
    MultiCall,
    ServerProxy,
    dumps,
    loads,
)
from xmlrpc.server import resolve_dotted_attribute

from aiohttp import web


class AsyncSimpleXMLRPCDispatcher:
    """
    Original not async version of this class is in the original python std lib:
    https://github.com/python/cpython/blob/main/Lib/xmlrpc/server.py.

    use_builtin_types=True allows the use of bytes-object which is preferred
    because else it's a custom xmlrpc.client.Binary which sucks.
    """

    def __init__(
        self, instance, allow_none=True, encoding="utf-8", use_builtin_types=True
    ):
        self.setup_rpc(allow_none=allow_none, encoding=encoding, use_builtin_types=True)
        self.register_instance(instance, True)

    def setup_rpc(self, allow_none=True, encoding="utf-8", use_builtin_types=True):
        self.funcs = {}
        self.instance = None
        self.allow_none = allow_none
        self.encoding = encoding or "utf-8"
        self.use_builtin_types = use_builtin_types

    def register_instance(self, instance, allow_dotted_names=True):
        self.instance = instance
        self.allow_dotted_names = allow_dotted_names
        self.register_multicall_functions()
        self.register_introspection_functions()

    def register_function(self, function=None, name=None):
        if function is None:
            return partial(self.register_function, name=name)

        if name is None:
            name = function.__name__
        self.funcs[name] = function

        return function

    def register_introspection_functions(self):
        self.funcs.update(
            {
                "system.listMethods": self.system_listMethods,
                "system.methodSignature": self.system_methodSignature,
                "system.methodHelp": self.system_methodHelp,
            }
        )

    def register_multicall_functions(self):
        self.funcs.update({"system.multicall": self.system_multicall})

    async def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        try:
            params, method = loads(data, use_builtin_types=self.use_builtin_types)

            if dispatch_method is not None:
                response = dispatch_method(method, params)
            else:
                response = await self._dispatch(method, params)
            response = (response,)
            response = dumps(
                response,
                methodresponse=1,
                allow_none=self.allow_none,
                encoding=self.encoding,
            )
        except Fault as fault:
            response = dumps(fault, allow_none=self.allow_none, encoding=self.encoding)
        except BaseException as exc:
            response = dumps(
                Fault(1, f"{type(exc)}:{exc}"),
                encoding=self.encoding,
                allow_none=self.allow_none,
            )

        return response.encode(self.encoding, "xmlcharrefreplace")

    def system_listMethods(self):
        methods = set(self.funcs.keys())
        if self.instance is not None:
            if hasattr(self.instance, "_listMethods"):
                methods |= set(self.instance._listMethods())
            elif not hasattr(self.instance, "_dispatch"):
                methods |= set(list_public_methods(self.instance))
        return sorted(methods)

    def system_methodSignature(self, method_name):
        return "signatures not supported"

    def system_methodHelp(self, method_name):
        method = None
        if method_name in self.funcs:
            method = self.funcs[method_name]
        elif self.instance is not None:
            if hasattr(self.instance, "_methodHelp"):
                return self.instance._methodHelp(method_name)
            elif not hasattr(self.instance, "_dispatch"):
                try:
                    method = resolve_dotted_attribute(
                        self.instance, method_name, self.allow_dotted_names
                    )
                except AttributeError:
                    pass
        if method is None:
            return ""
        else:
            return pydoc.getdoc(method)

    async def system_multicall(self, call_list):
        results = []
        for call in call_list:
            method_name = call["methodName"]
            params = call["params"]

            try:
                results.append([await self._dispatch(method_name, params)])
            except Fault as fault:
                results.append(
                    {"faultCode": fault.faultCode, "faultString": fault.faultString}
                )
            except BaseException as exc:
                results.append({"faultCode": 1, "faultString": f"{type(exc)}:{exc}"})
        return results

    async def _dispatch(self, method, params):
        try:
            func = self.funcs[method]
        except KeyError:
            pass
        else:
            if func is not None:
                return await func(*params)
            raise Exception(f'method "{method}" is not supported')

        if self.instance is not None:
            if hasattr(self.instance, "_dispatch"):
                return await self.instance._dispatch(method, params)

            try:
                func = resolve_dotted_attribute(
                    self.instance, method, self.allow_dotted_names
                )
            except AttributeError:
                pass
            else:
                if func is not None:
                    return await func(*params)

        raise Exception(f'method "{method}" is not supported')


def rpc_wrap_instance(obj):

    class Session:

        def __init__(self, data=None):
            self._data = data or {}

        async def get(self, key, default=None):
            return self._data.get(key, default)

        async def set(self, key, value):
            self._data[key] = value

        async def delete(self, key):
            try:
                del self._data[key]
                return True
            except KeyError:
                return False

        async def exists(self, key):
            return key in self._data

    class Instance:

        def __init__(self, _self):
            self._self = self
            self.session = Session()

        def __get__(self, key):
            return getattr(self._self, key)

        def ping(self, *args, **kwargs):
            return {"args": args, "kwargs": kwargs, "timestamp": str(datetime.now())}

    instance = Instance(obj)

    for attr in dir(obj):
        if attr == "rpc_handler":
            continue
        if attr.startswith("rpc_") and callable(getattr(obj, attr)):
            setattr(instance, attr[4:], getattr(obj, attr))

    return instance


class Application(web.Application):

    def __init__(self, url=None, host=None, port=None, *args, **kwargs):
        self.host = host
        self.port = port
        self._url = url
        self._rpc = None
        if self.rpc_url:
            self._rpc = ServerProxy(self.rpc_url)
        super().__init__(*args, **kwargs)
        self.arpc = rpc_wrap_instance(self)
        self.rpc_dispatcher = AsyncSimpleXMLRPCDispatcher(self.arpc)
        self.router.add_post("/rpc", self.rpc_handler)

    def __get__(self, key):
        if self._rpc:
            return getattr(self._rpc, key)
        return getattr(self.arpc, key)

    @property
    def url(self):
        if self._url:
            return self._url
        return f"http://{self.host}:{self.port}"

    @property
    def rpc_url(self):
        return self.url.rstrip("/") + "/rpc"

    def connect(self, url):
        return ServerProxy(url)

    def multicall(self, url):
        return MultiCall(self.connect(url))

    @property
    def rpc(self):
        if not self._rpc:
            self._rpc = ServerProxy(url or self.rpc_url)
        return self._rpc

    async def rpc_handler(self, request):
        request_body = await request.text()
        response_body = await self.rpc_dispatcher._marshaled_dispatch(request_body)
        return web.Response(text=response_body.decode())
